/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.field.shape;

import com.vividsolutions.jts.geom.Geometry;
import java.awt.datatransfer.Transferable;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.Action;
import nl.cloudfarming.client.field.model.Shape;
import nl.cloudfarming.client.field.model.ShapeCollection;
import nl.cloudfarming.client.field.model.ShapeImportSource;
import nl.cloudfarming.client.geoviewer.BaseLayer;
import nl.cloudfarming.client.geoviewer.LayerInfo;
import nl.cloudfarming.client.geoviewer.RandomColorPalette;
import nl.cloudfarming.client.geoviewer.SimpleLayer;
import org.geotools.data.FeatureSource;
import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.type.AttributeDescriptor;
import org.opengis.feature.type.AttributeType;
import org.opengis.feature.type.Name;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.ChildFactory;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.datatransfer.PasteType;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;

/**
 * DataObject for ESRI shape files
 *
 */
@NbBundle.Messages("ShapeDataObject.shape_name=Unknown")
public class ShapeDataObject extends MultiDataObject {

    public static final int CONTENT_NONE = 0;
    public static final int CONTENT_FIELDS = 1;
    public static final int CONTENT_TREATMENT_ZONES = 2;
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.shapetype.ShapeDataObject");
    private static final String ATT_NAME_PROPERTY = "att_name_property";
    private static final String ATT_VALUE_PROPERTY = "att_value_property";
    private static final String ATT_CONTENT_TYPE_PROPERTY = "att_content_type_property";
    public static final String MIMETYPE = "application/x-esri-shape";
    private final InstanceContent ic = new InstanceContent();
    private final Lookup lkp = new AbstractLookup(ic);
    private List<Shape> shapes = null;
    private List<String> stringAttributes;
    private List<String> integerAttributes;
    private final TreatmentZoneImportCookie treatmentZoneImportCookie = new TreatmentZoneImportCookie(this);
    private final FieldImportCookie fieldImportCookie = new FieldImportCookie(this);

    public ShapeDataObject(FileObject pf, MultiFileLoader loader) throws DataObjectExistsException, IOException {
        super(pf, loader);
        registerEditor(MIMETYPE, true);
        ic.add(new ShapeImportSource() {
            @Override
            public Set<FileObject> getFiles() {
                return files();
            }
        });
        checkDefaultAttributes();
        updateImportCookie();
    }

    // check shapefile attributes against registered property mappings:
    private void checkDefaultAttributes() {
        // only check for attributes when no configuration has been done yet:
        if(getContentType() == CONTENT_NONE && getNameProperty() == null && getValueProperty() == null) {
            updateMappingAttributes(ShapePropertyMappingRegistry.getInstance().findAll(), false);
        }
    }
    
    /**
     * Check shapefile attributes against a custom list of property mappings,
     * possibly clearing existing attributes when no match is found.
     * 
     * @param mappings 
     */
    public void updateMappingAttributes(List<ShapePropertyMapping> mappings, boolean reset) {

        if (reset) {
            setNameProperty(null);
            setValueProperty(null);
            setContentType(CONTENT_NONE);
            for (Shape shape : getShapes()){
                shape.setName(null);
                shape.setValue(null);
            }
        }

        if (mappings == null || mappings.isEmpty()) {
            return;
        }
            
        final List<String> nameProperties = new LinkedList<>(getStringAttributes());
        nameProperties.addAll(getIntegerAttributes());
        final List<String> valueProperties = new LinkedList<>(getIntegerAttributes());
        final ShapePropertyMapping mapping = ShapePropertyMappingRegistry.findMatch(mappings, nameProperties, valueProperties);
        if (mapping != null) {
            LOGGER.log(Level.INFO, "applying shape attribute mapping: {0}", mapping);
            setNameProperty(mapping.getNameProperty());
            setValueProperty(mapping.getValueProperty());
            setContentType(CONTENT_TREATMENT_ZONES);
            for (Shape shape : getShapes()){
                shape.setNameFromAttribute(mapping.getNameProperty());
                shape.setValueFromAttribute(mapping.getNameProperty());
            }
        }
    }
    
    /**
     * Method to set the secondary files
     *
     * @param secondary the secondary files
     */
    public void setSecondaryFiles(List<FileObject> secondary) {
        for (FileObject fileObject : secondary) {
            registerEntry(fileObject);
        }
    }

    @Override
    protected Node createNodeDelegate() {
        Node delegate = super.createNodeDelegate();
        ic.add(new ShapeCollection(getShapes()));
        //FIXME move palette to object level
        ic.add(new BaseLayer(new RandomColorPalette(),getPrimaryFile().getName()));
        LayerInfo layerInfo = new LayerInfo();
        layerInfo.setShowIcons(false);
        ic.add(layerInfo);
        return new ShapeDataObjectNode(delegate, getPrimaryFile().getName());
    }

    void setNameProperty(String nameProperty) {
        try {
            getPrimaryFile().setAttribute(ATT_NAME_PROPERTY, nameProperty);
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "could not save name propery mapping for shape data object {0}", getName());
        }
    }

    /**
     * Name of the attribute which value should be used for the name of the
     * shape
     *
     * @return
     */
    String getNameProperty() {
        return (String) getPrimaryFile().getAttribute(ATT_NAME_PROPERTY);
    }

    void setValueProperty(String valueProperty) {
        try {
            getPrimaryFile().setAttribute(ATT_VALUE_PROPERTY, valueProperty);
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "could not save value propery mapping for shape data object {0}", getName());
        }
    }

    /**
     * Retrieve the value property for a shapefile. The value property is name
     * of the attribute in the shape file use to map its value to the value
     * property of the shape.
     *
     * @return
     */
    String getValueProperty() {
        return (String) getPrimaryFile().getAttribute(ATT_VALUE_PROPERTY);
    }

    void setContentType(int contentType) {
        try {
            getPrimaryFile().setAttribute(ATT_CONTENT_TYPE_PROPERTY, contentType);
            updateImportCookie();
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "could not save content type propery mapping for shape data object {0}", getName());
        }
    }

    private void updateImportCookie() {
        int contentType = getContentType();

        if (CONTENT_TREATMENT_ZONES == contentType && getLookup().lookup(TreatmentZoneImportCookie.class) == null) {
            ic.add(treatmentZoneImportCookie);
        } else if (CONTENT_TREATMENT_ZONES != contentType && getLookup().lookup(TreatmentZoneImportCookie.class) != null) {
            ic.remove(treatmentZoneImportCookie);
        }

        if (CONTENT_FIELDS == contentType && getLookup().lookup(FieldImportCookie.class) == null) {
            ic.add(fieldImportCookie);
        } else if (CONTENT_FIELDS != contentType && getLookup().lookup(FieldImportCookie.class) != null) {
            ic.remove(fieldImportCookie);
        }
    }

    /**
     * Get the content type for this shapefile <p>possible values: <ul>
     * <li>{@link ShapeDataObject#CONTENT_NONE}</li>
     * <li>{@link ShapeDataObject#CONTENT_FIELDS}</li>
     * <li>{@link ShapeDataObject#CONTENT_TREATMENT_ZONES}</li> <ul>
     *
     * </p>
     *
     * @return
     */
    public int getContentType() {
        Integer contentType = (Integer) getPrimaryFile().getAttribute(ATT_CONTENT_TYPE_PROPERTY);
        return contentType == null ? CONTENT_NONE : contentType.intValue();
    }

    @Override
    public Lookup getLookup() {
        return new ProxyLookup(super.getLookup(), lkp);
    }

    /**
     * Retrieve all shapes in this shapeDataObject
     *
     * @return
     */
    public final List<Shape> getShapes() {
        if (shapes == null) {
            shapes = new ArrayList<>();
            try {
                ShapefileDataStore shpDataStore = new ShapefileDataStore(getPrimaryFile().getURL());
                String typeName = shpDataStore.getTypeNames()[0];
                FeatureSource featureSource = shpDataStore.getFeatureSource(typeName);

                FeatureCollection result = featureSource.getFeatures();
                FeatureIterator iterator = result.features();
                while (iterator.hasNext()) {
                    SimpleFeature feature = (SimpleFeature) iterator.next();
                    Shape shape = getShape(feature);
                    //
                    // parse all available string attributes
                    //
                    stringAttributes = readAttributes(feature, String.class);
                    //
                    // add all string attributes and their values to the map in the shape object
                    //
                    for (String att : stringAttributes) {
                        int index = feature.getFeatureType().indexOf(att);
                        if (index >= 0) {
                            String value = (String) feature.getAttribute(index);
                            shape.addStringAttribute(att, value);
                        }
                    }
                    //
                    // parse all available Integer attributes
                    //
                    integerAttributes = readAttributes(feature, Integer.class);
                    //
                    // add all string attributes and their values to the map in the shape object
                    //
                    for (String att : integerAttributes) {
                        int index = feature.getFeatureType().indexOf(att);
                        if (index >= 0) {
                            Integer value = (Integer) feature.getAttribute(index);
                            shape.addIntegerAttribute(att, value == null ? null : value.doubleValue());
                        }
                    }

                    List<String> longValues = readAttributes(feature, Long.class);
                    for (String att : longValues) {
                        int index = feature.getFeatureType().indexOf(att);
                        if (index >= 0) {
                            Long value = (Long) feature.getAttribute(index);
                            shape.addIntegerAttribute(att, value == null ? null : value.doubleValue());
                        }
                    }
                    integerAttributes.addAll(longValues);

                    List<String> doubleValues = readAttributes(feature, Double.class);
                    for (String att : doubleValues) {
                        int index = feature.getFeatureType().indexOf(att);
                        if (index >= 0) {
                            Double value = (Double) feature.getAttribute(index);
                            shape.addIntegerAttribute(att, value);
                        }
                    }
                    integerAttributes.addAll(doubleValues);
                    shape.setNameFromAttribute(getNameProperty());
                    shape.setValueFromAttribute(getValueProperty());

                    shapes.add(shape);
                }
                iterator.close();
            } catch (IOException ex) {
                LOGGER.log(Level.WARNING, "Could not return shapes", ex.getMessage());
            }
        }
        return shapes;
    }

    /**
     * Reads attributes of type String from specified feature into the
     * attributes List
     *
     * @param feature
     */
    private List<String> readAttributes(SimpleFeature feature, Class clazz) {
        List<String> returnList = new ArrayList<>();
        List<AttributeDescriptor> attributeDescriptors = feature.getFeatureType().getAttributeDescriptors();
        for (AttributeDescriptor attributeDescriptor : attributeDescriptors) {
            // Determine type
            AttributeType attributeType = attributeDescriptor.getType();

            boolean attributeIsString = attributeType.getBinding().equals(clazz);

            if (attributeIsString) {
                // Determine Attributename
                Name name = attributeDescriptor.getName();
                if (name != null) {
                    returnList.add(name.toString());
                }
            }
        }
        return returnList;
    }

    /**
     * Get the attributes of type 'Integer' from this shp-file
     *
     * @return The attributes inside the shp-file
     */
    public List<String> getIntegerAttributes() {
        if (integerAttributes == null) {
            getShapes();
        }
        return integerAttributes;
    }

    /**
     * Get the attributes of type 'String' from this shp-file
     *
     * @return The attributes inside the shp-file
     */
    public List<String> getStringAttributes() {
        // Read shapes if not already done
        if (stringAttributes == null) {
            getShapes();
        }
        return stringAttributes;
    }

    /**
     *
     * @param payload
     * @return
     */
    private Shape getShape(SimpleFeature payload) {

        int attIndexPlotPoly = payload.getFeatureType().indexOf("the_geom");

        Geometry geometry = null;
        if (attIndexPlotPoly >= 0) {
            geometry = (Geometry) payload.getAttribute(attIndexPlotPoly);
        }
        String shapeName = Bundle.ShapeDataObject_shape_name();

        Shape shape = new Shape(geometry);
        shape.setName(shapeName);
        return shape;
    }

    @Override
    protected int associateLookup() {
        return 1;
    }

    /**
     * Root node for the shapefile
     */
    private final class ShapeDataObjectNode extends AbstractNode {

        final Node original;

        public ShapeDataObjectNode(Node original, String name) {
            super(Children.create(new ShapeChildFactory(), true), original.getLookup());
            this.original = original;
            setIconBaseWithExtension("nl/cloudfarming/client/field/shape/file_extension_shp.png");
            setDisplayName(name);
        }

        @Override
        public Action[] getActions(boolean context) {
            return original.getActions(context);
        }

        @Override
        public PasteType getDropType(Transferable t, int action, int index) {
            return null;
        }

        @Override
        public boolean canDestroy() {
            return original.canDestroy();
        }

        @Override
        public void destroy() throws IOException {
            original.destroy();
        }
    }

    /**
     * Creates shape nodes for the simple features within the specified
     * shapefile
     */
    private final class ShapeChildFactory extends ChildFactory<Shape> {

        @Override
        protected boolean createKeys(List<Shape> toPopulate) {
            // Add all features in this Dataobject to the nodelist
            toPopulate.addAll(ShapeDataObject.this.getShapes());
            return true;
        }

        @Override
        protected Node createNodeForKey(Shape key) {
            Node shapeNode = new ShapeNode(key);
            return shapeNode;
        }
    }

    /**
     * Node representing a SimpleFeature (shape) in a shape file
     */
    private final class ShapeNode extends AbstractNode {

        public ShapeNode(Shape shape) {
            // AGROSENSE-1163: add the containing data object to the lookup to link the shape to a project
            super(Children.LEAF, Lookups.fixed(new ShapeCollection(shape), shape, 
                    new RandomColorPalette(), new SimpleLayer(shape.getName()), ShapeDataObject.this));
            setName(shape.getName());
            shape.addPropertyChangeListener(new PropertyChangeListener() {
                @Override
                public void propertyChange(PropertyChangeEvent evt) {
                    if (Shape.PROP_NAME.equals(evt.getPropertyName())) {
                        setName((String) evt.getNewValue());
                    }
                }
            });
            setIconBaseWithExtension(shape.getIconPath());
        }

        @Override
        public PasteType getDropType(Transferable t, int action, int index) {
            return null;
        }
    }
}
